import importlib

class Engine(object):
	def __init__(self,settings,autologin=True):
		self._exceptions = importlib.import_module("pydirectory.%(type)s.exceptions" % {'type':settings.type})
		self._settings = settings
		if autologin:
			self.login()

	def authenticate(self,username,password):
		return False

	def login(self):
		return False



# -*- coding: utf-8 -*-
class attribute(object):
	_is_readonly = False
	def __init__(self,value,objects,objid=None,modify=True):
		self._id = objid
		self._objects = objects
		self._exceptions = self._objects._exceptions
		self._is_modified = False
		self._is_append = False
		self._is_delete = False
		self._delattr = []
		self._addattr = []
		self._toraw(value)

		if modify:
			self._is_modified = True
		else:
			self._is_modified = False

	def _toraw(self,value):
		if not (type(value) == list):
			self._raw = [value]
		elif value == None:
			self._raw = [None]
		else:
			self._raw = value

	def __str__(self):
		if (type(self._tohuman()) == str):
			return self._tohuman()
		else:
			return str(self._tohuman())

	def __unicode__(self): #fix python 2.7 compatibility
		if (type(self._tohuman()) == str):
			return self._tohuman().decode('utf-8')
		else:
			return unicode(self._tohuman())

	def __repr__(self):
		if type(self._tohuman()) == str:
			return self._tohuman()
		else:
			return repr(self._tohuman())

	def __setitem__(self,key,value):
		if self._is_readonly:
			raise self._exceptions.IsReadOnly
		self._raw[key] = value

	def __getitem__(self,key):
		return self.raw[key]

	def __len__(self):
		return len(self.raw)

	def __delitem__(self,key):
		if self._is_readonly:
			raise self._exceptions.IsReadOnly
		if self._is_modified:
			raise self._exceptions.ObjectIsModified
		value = self._raw[key]
		changed = False
		if value in self._addattr:
			key = self._addattr.index(value)
			del self._addattr[key]
			changed = True
		if not value in self._delattr:
			self._delattr.append(value)
			changed = True
		if changed:
			self._is_delete = True

	def __iter__(self):
		return iter(self._raw)

	def append(self,value):
		if self._is_modified:
			raise self._exceptions.ObjectIsModified
		changed = False
		if value in self._delattr:
			key = self._delattr.index(value)
			del self._delattr[key]
			changed = True
		if not value in self._addattr:
			self._addattr.append(value)
			changed = True
		if changed:
			self._is_append = True

	def update(self,value):
		if type(value) != list:
			self._raw = [value]
		else:
			self._raw = value

	def _tohuman(self):
		return self.value

	@property
	def value(self):
		return self._tovalue()

	def _tovalue(self):
		try: #Fix to have python 2.x and python 3.x compatibility
			typestr = unicode #On python 3 unicode type case not exist
			unicodebase = True #is python 2
		except:
			typestr = str
			unicodebase = False #is python 3
		result = self.raw
		if len(self._raw) == 1:
			result = self._raw[0]
			if (type(self._raw[0]) == bytes) or (type(self._raw[0]) == str) or (type(self._raw[0]) == typestr):
				if not unicodebase:
					if type(result) == bytes:
						result = result.decode('utf-8') #Fix to have python 2.x and python 3.x compatibility

		return result

	@property
	def raw(self):
		return self._raw



from ldap3 import Server, Connection, ALL, SUBTREE, BASE
from pydirectory.directory.engine.classes import Engine

class engine (Engine):

	def _authenticate(self,username,password,login=False):
		server_options = {
		 'host': repr(self._settings.host),
		 'use_ssl': self._settings.ssl.value,
		 'get_info': ALL
		}
		if self._settings.port.value != None:
			server_options['port'] = self._settings.port.value

		server = Server(**server_options)
		conn = Connection(server, user=username,password=password,auto_referrals=False,raise_exceptions=True)
		result = False
		try:
			result = conn.bind()
		except self._exceptions.LDAPSocketOpenError:
			raise self._exceptions.InvalidServer
		except self._exceptions.LDAPInvalidCredentialsResult:
			raise self._exceptions.InvalidCredentials


		if self._settings.port.value == None:
			self._settings.port = conn.server.port

		if login:
			if result == False:
				raise self._exceptions.InvalidCredentials
			if self._settings.basedn == None:
				self._settings.basedn = conn.server.info.raw['defaultNamingContext'][0].decode('utf-8') #fix to convert byte to string
			return conn
		return result


	def login(self):
		self._worker = self._authenticate(repr(self._settings.username),repr(self._settings.password),login=True)

	def authenticate(self,username,password):
		return self._authenticate(username,password)



from pydirectory.ldap.objects.attributes import *

class PrimaryGroupID(attribute):
	def _tovalue(self):
		sid = '%(base)s-%(id)s' % {
			'base':self._objects.base.objectsid.value,
			'id':int(self.raw[0])
		}
		return self._objects.get(self._objects.setQuery(objectsid=sid))

class member(attribute):
	_is_readonly = True

	@property
	def raw(self):
		if self._raw[0] == None:
			member = []
			init = 0
			for end in range(1500,10000,1500):
				try:
					fieldname = 'member;range=%(init)s-%(end)s' % {"init":init,"end":end-1}
					fieldwildcard = 'member;range=%(init)s-*' % {"init":init}
					result = self._objects.search(dn=self._id,attributes=[fieldname])[0]
					if fieldname in result.attributes.keys():
						member.extend(result.attributes[fieldname])
					if fieldwildcard in result.attributes.keys():
						member.extend(result.attributes[fieldwildcard])
					init=end
				except self._exceptions.LDAPOperationsErrorResult:
					self._raw = member

		return self._raw

class memberOf(attribute):
	_is_readonly = True

class userAccountControl(attribute):
	types = {
		"SCRIPT" : 1,
		"ACCOUNTDISABLE" : 2,
		"HOMEDIR_REQUIRED" : 8,
		"LOCKOUT" : 16,
		"PASSWD_NOTREQD" : 32,
		"ENCRYPTED_TEXT_PWD_ALLOWED" : 128,
		"TEMP_DUPLICATE_ACCOUNT" : 256,
		"NORMAL_ACCOUNT" : 512,
		"INTERDOMAIN_TRUST_ACCOUNT" : 2048,
		"WORKSTATION_TRUST_ACCOUNT" : 4096,
		"SERVER_TRUST_ACCOUNT" : 8192,
		"DONT_EXPIRE_PASSWORD" : 65536,
		"MNS_LOGON_ACCOUNT" : 131072,
		"SMARTCARD_REQUIRED" : 262144,
		"TRUSTED_FOR_DELEGATION" : 524288,
		"NOT_DELEGATED" : 1048576,
		"USE_DES_KEY_ONLY" : 2097152,
		"DONT_REQ_PREAUTH" : 4194304,
		"PASSWORD_EXPIRED" : 8388608,
		"TRUSTED_TO_AUTH_FOR_DELEGATION" : 16777216
	}

	def accountControl(self,options):
		val = 0
		for type in options:
			val = val + self.types.get(type,0)
		return val

	def _toraw(self,value):
		if type(value) == list:
			acControl = self.accountControl(value)
			if acControl > 0:
				self._raw = [str(acControl)]
			else:
				self._raw = value
		else:
			self._raw = [value]

	def _tohuman(self):
		try:
			binary = bin(int(self.value))[2:][::-1]
		except:
			return self.value
		values = []
		for bit in range(0,len(binary)):
			if binary[bit] == '1':
				decimal = int(binary[bit]+'0'*bit,2)
				try:
					position = list(self.types.values()).index(decimal)
					value = list(self.types.keys())[position]
					values.append(value)
				except:
					values.append("UNKNOWN")
		return values


class unicodePwd(attribute):
	def _toraw(self,value):
		try:
			password = ('"%s"' % value).encode('utf-16-le')
		except UnicodeDecodeError:
			password = ('"%s"' % value.decode('utf-8')).encode('utf-16-le')
		self._raw = [password]

	def _tohuman(self):
		raise self._exceptions.PasswordPrintNotAllowed


class ObjectSid(attribute):
	_is_readonly = True

	def _littleEndian(self,hex):
		result = '';
		xinit = len(hex) - 2
		for x in range(xinit,0-1,-2):
			result += hex[x:x+2]
		return result

	def _getTextSID(self,binsid):
		import binascii
		hex_sid = binascii.hexlify(binsid).decode('utf-8')
		subcount = int(hex_sid[2:2+2],16)
		rev = int(hex_sid[0:0+2],16)
		auth = int(hex_sid[4:4+12],16)
		result = str(rev)+'-'+str(auth)
		subauth = {}
		for x in range(0,subcount):
			le = hex_sid[16+(x*8):(16+(x*8))+8]
			subauth[x] = int(self._littleEndian(le),16)
			try:
				result += '-'+str(subauth[x])
			except TypeError:
				result += '-'+bytes(subauth[x])
		return 'S-'+result

	def _tovalue(self):
		return self._getTextSID(self.raw[0])


class sIDHistory(ObjectSid):
	pass


class ObjectGUID(attribute):
	_is_readonly = True
	def _tovalue(self):
		import uuid
		return str(uuid.UUID(bytes=self.raw[0]))


class formattime(attribute):
	_is_readonly = True
	def _tovalue(self):
		#Hack from ldap3.protocol.formatters.formatters.format_time https://github.com/cannatag/ldap3
		from pydirectory.activedirectory.tools import OffsetTzInfo
		from datetime import datetime
		raw_value = self.raw[0]
		if len(raw_value) < 10 or not all((c in b'0123456789+-,.Z' for c in raw_value)) or (b'Z' in raw_value and not raw_value.endswith(b'Z')):  # first ten characters are mandatory and must be numeric or timezone or fraction
			return raw_value

		year = int(raw_value[0: 4])
		month = int(raw_value[4: 6])
		day = int(raw_value[6: 8])
		hour = int(raw_value[8: 10])
		minute = 0
		second = 0
		microsecond = 0

		remain = raw_value[10:]
		if remain and remain.endswith(b'Z'):  # uppercase 'Z'
			sep = b'Z'
		elif b'+' in remain:  # timezone can be specified with +hh[mm] or -hh[mm]
			sep = b'+'
		elif b'-' in remain:
			sep = b'-'
		else:  # timezone not specified
			return raw_value

		time, _, offset = remain.partition(sep)

		if time and (b'.' in time or b',' in time):
			# fraction time
			if time[0] in b',.':
				minute = 6 * int(time[1] if str == bytes else chr(time[1]))
			elif time[2] in b',.':
				minute = int(raw_value[10: 12])
				second = 6 * int(time[3] if str == bytes else chr(time[3]))
			elif time[4] in b',.':
				minute = int(raw_value[10: 12])
				second = int(raw_value[12: 14])
				microsecond = 100000 * int(time[5] if str == bytes else chr(time[5]))
			elif len(time) == 2:  # mmZ format
				minute = int(raw_value[10: 12])
			elif len(remain) == 0:  # Z format
				pass
			elif len(time) == 4:  # mmssZ
				minute = int(raw_value[10: 12])
				second = int(raw_value[12: 14])
			else:
				return raw_value

			if sep == b'Z':  # UTC
				timezone = OffsetTzInfo.OffsetTzInfo(0, 'UTC')
			else:  # build timezone
				try:
					if len(offset) == 2:
						timezone_hour = int(offset[:2])
						timezone_minute = 0
					elif len(offset) == 4:
						timezone_hour = int(offset[:2])
						timezone_minute = int(offset[2:4])
					else:  # malformed timezone
						raise ValueError
				except ValueError:
					return raw_value
				if str != bytes:  # python3
					timezone = OffsetTzInfo.OffsetTzInfo((timezone_hour * 60 + timezone_minute) * (1 if sep == b'+' else -1), 'UTC' + str(sep + offset, encoding='utf-8'))
				else:
					timezone = OffsetTzInfo.OffsetTzInfo((timezone_hour * 60 + timezone_minute) * (1 if sep == b'+' else -1), unicode('UTC' + sep + offset, encoding='utf-8'))

			try:
				return datetime(year=year,
								month=month,
								day=day,
								hour=hour,
								minute=minute,
								second=second,
								microsecond=microsecond,
								tzinfo=timezone)
			except (TypeError, ValueError):
				return raw_value


class WhenCreated(formattime):
	pass

class WhenChanged(formattime):
	pass

class dSCorePropagationData(formattime):
	pass

class adTimeStamp(attribute):
	_is_readonly = True
	def _tovalue(self):
		#Hack from ldap3.protocol.formatters.formatters.format_ad_timestamp https://github.com/cannatag/ldap3
		from pydirectory.activedirectory.tools import OffsetTzInfo
		from datetime import datetime
		try:
			timestamp = int(self.raw[0])
			return datetime.fromtimestamp(timestamp / 10000000.0 - 11644473600, tz=OffsetTzInfo.OffsetTzInfo(0, 'UTC'))
		except Exception:
			if self.raw[0] == b'9223372036854775807':
				return datetime.max
			return self.raw[0]

class creationTime(adTimeStamp):
	pass

class pwdLastSet(adTimeStamp):
	pass

class badPasswordTime(adTimeStamp):
	pass

class lastLogon(adTimeStamp):
	pass

class lastLogonTimeStamp(adTimeStamp):
	pass

class accountExpires(adTimeStamp):
	pass

class maxPwdAge(adTimeStamp):
	pass

class minPwdAge(adTimeStamp):
	pass



import importlib
import pydirectory.directory.objects.types


class objecttypes(object):
	def __init__(self,objects):
		self._objects = objects

	def _getobject(self,value):
		return importlib.import_module("pydirectory.%(type)s.objects.types" % {'type':self._objects._engine._settings.type}).object(self._objects,value)

	def __call__(self,value):
		return self._getobject(value)

class objectslist(object):
	def __init__(self,objects): #change engine by object
		self._objects = objects
		self.objecttypes = importlib.import_module("pydirectory.%(type)s.objects.classes" % {'type':self._objects._engine._settings.type}).objecttypes(self._objects)
		self._store = []

	def __getitem__(self,key):
		return self._store[key]

	def __setitem__(self,key,data):
		self._store[key] = self._setobject(data)

	def __delitem__(self,key):
		del self._store[key]

	def __iter__(self):
		return iter(self._store)

	def __str__(self):
		return str(self._store)

	def __len__(self):
		return len(self._store)

	def __repr__(self):
		return repr(self._store)

	def _setobject(self,data):
		return self.objecttypes(data)

	def append(self,data):
		self._store.append(self._setobject(data))

class objects(object):
	def __init__(self,engine):
		self._engine = engine
		self._exceptions = importlib.import_module("pydirectory.%(type)s.exceptions" % {'type':self._engine._settings.type})
		self._search = importlib.import_module("pydirectory.%(type)s.objects.actions" % {'type':self._engine._settings.type}).search
		self._get = importlib.import_module("pydirectory.%(type)s.objects.actions" % {'type':self._engine._settings.type}).get
		self._new = importlib.import_module("pydirectory.%(type)s.objects.actions" % {'type':self._engine._settings.type}).new

	@property
	def search(self):
		return self._search(self)

	@property
	def get(self):
		return self._get(self)

	@property
	def new(self):
		return self._new(self)



class customException(Exception):
	def __init__(self,message=False):
		if message:
			self.message=message

	def __str__(self):
		return repr(self.message)


class NotValidController(customException):
	message = "Not Valid Controller"

class InvalidServer(customException):
	message = "Invalid Host or Server Name"

class UnknownError(customException):
	message = "Unknown Error"

class InvalidCredentials(customException):
	message = "Invalid username y/or password"

class AttributeNotValid(customException):
	message = "The attribute is not valid"

class SettingsOptionNotExist(customException):
	message = "The Settings Option to set not exist"

class InvalidValueType(customException):
	message = "Invalid value type"

class InvalidValue(customException):
	message = "Invalid value"

class MustBeOptionType(customException):
	message = "Value must be Option type"

class PasswordPrintNotAllowed(customException):
	message = "Password Print is not Allowed"

class DNSHostnameCanNotBeResolved(customException):
	message = "DNS Hostname Can not be resolved"

class MultipleResults(customException):
	message = "The response has returned multiple results"

class ObjectIsModified(customException):
	message = "The object is modified. You can not delete or append attributes values"

class ObjectNotExist(customException):
	message = "The object not exist"

class AttributeisReadOnly(customException):
	message = "The attribute is read only"

class CheckValueAttributes(customException):
	message = "Some attribute value is not correct"

class IsReadOnly(customException):
	message = "The action is not permitted. It is read only"



from pydirectory.ldap.tools import *
#from datetime import timedelta, tzinf

class setQuery(setQuery):
	specialFields = {
		'is_disable' : {
	 		True: '(userAccountControl:1.2.840.113556.1.4.803:=2)',
			False: '(!(userAccountControl:1.2.840.113556.1.4.803:=2))',
		},
		'is_enable' : {
	 		True: '(!(userAccountControl:1.2.840.113556.1.4.803:=2))',
			False: '(userAccountControl:1.2.840.113556.1.4.803:=2)',
		},
		'is_member': '(memberof:1.2.840.113556.1.4.1941:={DN})',
		'in_group': '(member:1.2.840.113556.1.4.1941:={DN})'
	}

# from python standard library docs
class OffsetTzInfo(object):
	from datetime import tzinfo
	class OffsetTzInfo(tzinfo):
		"""Fixed offset in minutes east from UTC"""

		def __init__(self, offset, name):
			import datetime
			self.datetime = datetime
			self.offset = offset
			self.name = name
			self._offset = self.datetime.timedelta(minutes=offset)
			self.name = name

		def __str__(self):
			return self.name

		def __repr__(self):

			return 'OffsetTzInfo(offset={0.offset!r}, name={0.name!r})'.format(self)

		def utcoffset(self, dt):
			return self._offset

		def tzname(self, dt):
			return self.name

		def dst(self, dt):
			return self.datetime.timedelta(0)

	def __init__(self,*args,**kwargs):
		pass



from pydirectory.ldap.objects.actions import *

class search (search):
	def users(self,query=None,*args,**kwargs):
		if query != None:
			query = '(&{query}(&(objectclass=top)(objectclass=person)(objectclass=organizationalPerson)(objectclass=user)(!(objectclass=computer))))'.format(query=query)
		return self.__call__(query,*args,**kwargs)

	def groups(self,query=None,*args,**kwargs):
		if query != None:
			query = '(&{query}(&(objectclass=top)(objectclass=group)))'.format(query=query)
		return self.__call__(query,*args,**kwargs)

	def computers(self,query=None,*args,**kwargs):
		if query != None:
			query = '(&{query}(&(objectclass=top)(objectclass=person)(objectclass=organizationalPerson)(objectclass=user)(objectclass=computer)))'.format(query=query)
		return self.__call__(query,*args,**kwargs)

	def ous(self,query=None,*args,**kwargs):
		if query != None:
			query = '(&{query}(&(objectclass=top)(objectclass=organizationalUnit)))'.format(query=query)
		return self.__call__(query,*args,**kwargs)

class get(get):
	def user(self,query=None,*args,**kwargs):
		if query != None:
			query = '(&{query}(&(objectclass=top)(objectclass=person)(objectclass=organizationalPerson)(objectclass=user)(!(objectclass=computer))))'.format(query=query)
		return self.__call__(query,*args,**kwargs)

	def group(self,query=None,*args,**kwargs):
		if query != None:
			query = '(&{query}(&(objectclass=top)(objectclass=group)))'.format(query=query)
		return self.__call__(query,*args,**kwargs)

	def computer(self,query=None,*args,**kwargs):
		if query != None:
			query = '(&{query}(&(objectclass=top)(objectclass=person)(objectclass=organizationalPerson)(objectclass=user)(objectclass=computer)))'.format(query=query)
		return self.__call__(query,*args,**kwargs)

	def ou(self,query=None,*args,**kwargs):
		if query != None:
			query = '(&{query}(&(objectclass=top)(objectclass=organizationalUnit)))'.format(query=query)
		if kwargs.get('dn',False):
			kwargs['scope'] = 'BASE'
		return self.__call__(query,*args,**kwargs)

class new (new):
	def user(self):
		data = {
			"dn":None,
			"container":None,
			"cn":None,
			"samaccountname":None,
			"objectClass":['top','person', 'organizationalPerson', 'user'],
			'useraccountcontrol': ['ACCOUNTDISABLE', 'NORMAL_ACCOUNT'],
			'userprincipalname': None,
			'givenname':None,
			'sn':None,
			'displayname':None,
		}
		return self._get(data)

	def group(self):
		data = {
			"dn":None,
			"container":None,
			"cn":None,
			"samaccountname":514,
			"objectClass":['top', 'group'],
		}
		return self._get(data)

	def computer(self):
		data = {
			"dn":None,
			"container":None,
			"cn":None,
			"samaccountname":None,
			"objectClass":['top','person', 'organizationalPerson', 'user', 'computer'],
			'useraccountcontrol': ['PASSWD_NOTREQD', 'WORKSTATION_TRUST_ACCOUNT'],
		}
		return self._get(data)


	def ou(self):
		data = {
			"dn":None,
			"container":None,
			"ou":None,
			"objectClass":['top', 'organizationalUnit'],
		}
		return self._get(data)



import importlib

class search(object):
	def __init__(self,objects):
		self._objects = objects
		self._exceptions = self._objects._exceptions
		self._objectslist = importlib.import_module("pydirectory.%(type)s.objects.classes" % {'type':self._objects._engine._settings.type}).objectslist(self._objects)

	def __call__(self,query=None,**kwargs):
		return self._get(query,**kwargs)

class new(object):
	def __init__(self,objects):
		self._objects = objects
		self._objectslist = importlib.import_module("pydirectory.%(type)s.objects.classes" % {'type':self._objects._engine._settings.type}).objectslist(self._objects)

	def __call__(self):
		return self._get()



import socket
import getpass
import importlib
from pydirectory.directory import exceptions

class Option(object):
	def __init__(self,null=True,value=None,is_password=False,is_hostname=False,type=None,validator=None,description=None):
		self.null = null
		self.__type = type
		self.is_password=is_password
		self.is_hostname=is_hostname
		self.__validator = validator
		self.__doc__ = description
		if value == None:
			if not self.null:
				self._input_text()
			else:
				self.value = value
		else:
			self.value = value

	def _input_text(self):
		if self.is_password:
			self.value=getpass.getpass(self.__doc__+': ')
		else:
			self.value=input(self.__doc__+': ')


	def __str__(self):
		result = str(self.value)
		if self.is_password:
			raise exceptions.PasswordPrintNotAllowed
		return result

	def __eq__ (self,other):
		return self.value == other

	def __repr__(self):
		if type(self.value) == str:
			return self.value
		return repr(self.value)

	def __setattr__(self,name,value):
		if name == "value":
			if value != None:
				try:
					value = self.__type(value)
				except ValueError as e:
					raise exceptions.InvalidValueType(e)

				if self.__validator != None:
					if not self.__validator(value):
						raise exceptions.InvalidValue

			if self.is_hostname:
				try:
					self.ipv4 = socket.getaddrinfo(value,None,socket.AF_INET)[0][4][0]
				except socket.gaierror:
					self.ipv4 = False

				try:
					self.ipv6 = socket.getaddrinfo(value,None,socket.AF_INET6)[0][4][0]
				except socket.gaierror:
					self.ipv6 = False

				if (not self.ipv4) and (not self.ipv6):
					raise exceptions.DNSHostnameCanNotBeResolved

		super(Option,self).__setattr__(name,value)


class Settings(object):

	def __init__(self,type,**kwargs):
		self._exceptions = importlib.import_module("pydirectory.%(type)s.exceptions" % {'type':type})
		self.type = Option(null=False,type=str,description="Type engine",value=type)
		self._options(**kwargs)

	def __setattr__(self,name,value):
		if self.__dict__.get(name,False):
			self.__dict__[name].value = value
		else:
			if (type(value) != Option) and (name.find('_') != 0):
				raise self._exceptions.MustBeOptionType
			super(Settings,self).__setattr__(name,value)



from pydirectory.directory.settings.classes import Settings, Option

class settings(Settings):
	def _options(self,**kwargs):
		self.host = Option(null=False,type=str,is_hostname=True,description="Domain Controler Hostname or IP",value=kwargs.get('hostname',None))
		self.port = Option(type=int,description="Domain Controler Port connection",value=kwargs.get('port',None))
		self.username = Option(type=str,description="Username user connection",value=kwargs.get('username',None))
		passnull = True
		if self.username != None:
			passnull = False
		self.password = Option(null=passnull,type=str,is_password=True,description="Password user connection",value=kwargs.get('password',None))
		self.basedn= Option(type=str,description="LDAP base DN",value=kwargs.get('basedn',None))
		self.ssl= Option(type=bool,description="Enable SSL Connection (True/False)",value=kwargs.get('ssl',None))
		self.tls= Option(type=bool,description="Enable TLS Connection (True/False)",value=kwargs.get('tls',None))



class setQuery(object):
	specialFields = {}
	operators = {
		'default' : '({field}={value})',
		'conditions': {
			'gt' : '({field}>{value})',
			'ge' : '({field}>={value})',
			'lt' : '({field}<{value})',
			'le' : '({field}<={value})',
			'eq' : '({field}={value})',
		},
		'wrapper' : {
			'not' : '(!{query})'
		}
	}

	def __init__(self,*args,**kwargs):
		pass

	def __call__(self,oper='and',**fields):
		result = ''
		for field,value in fields.items():
			if field in self.specialFields.keys():
				if type(self.specialFields[field]) == dict:
					result += self.specialFields[field][value]
				else:
					if type(value) == list:
						for pos in value:
							result += self.specialFields[field].format(DN=pos)
					else:
						result += self.specialFields[field].format(DN=value)
				continue

			wrapper = False
			for wrap in self.operators['wrapper'].keys():
				if field.find('__'+wrap) > -1:
					field = field.replace('__'+wrap,'')
					wrapper = self.operators['wrapper'][wrap]

			try:
				name,operator = field.split('__')
			except ValueError:
				name = field
				operator = None

			if type(value) == list:
				query = ''
				for val in value:
					aux = self.operators['conditions'].get(operator,self.operators['default']).format(field=name,value=val)
					if wrapper:
						aux = wrapper.format(query=aux)
					query += aux
			else:
				query = self.operators['conditions'].get(operator,self.operators['default']).format(field=name,value=value)
				if wrapper:
					query = wrapper.format(query=query)

			result += query

		if oper == 'and':
			return '(&{query})'.format(query=result)
		if oper == 'or'	:
			return '(|{query})'.format(query=result)



import importlib


class Directory(object):
	def __init__(self,type,autologin=True,**kwargs):
		self.settings = importlib.import_module("pydirectory.%(type)s.settings.classes" % {'type':type}).settings(type,**kwargs)
		self.engine = importlib.import_module("pydirectory.%(type)s.engine.classes" % {'type':type}).engine(self.settings,autologin)
		self.objects = importlib.import_module("pydirectory.%(type)s.objects.classes" % {'type':type}).objects(self.engine)

		tools = importlib.import_module("pydirectory.%(type)s.tools" % {'type':type})
		for tool in dir(tools):
			if tool.find('_') != 0:
				module = getattr(tools,tool)(self.objects)
				setattr(self,tool,module)
				setattr(self.objects,tool,module)



from pydirectory.directory.exceptions import *
from ldap3 import LDAPSocketOpenError,LDAPInvalidFilterError,LDAPChangesError,LDAPInvalidCredentialsResult,LDAPSocketSendError,LDAPOperationsErrorResult,LDAPEntryAlreadyExistsResult,LDAPUnwillingToPerformResult, LDAPNoSuchObjectResult
from pyasn1.error import PyAsn1Error

class LDAPError(customException):
	pass


class LDAPReferrals(LDAPError):
	pass

class DNisNone(customException):
	message = "DN is None"



from pydirectory.ldap.objects import types

class object(types.object):
	_type = {}

class user(object):
	_type = {
		'objectClass' : [b'top', b'person', b'organizationalPerson', b'user']
	}
	@property
	def is_enable(self):
		mod = int(self.useraccountcontrol.value) % 8
		if mod == 0:
			return True
		else:
			return False

	@property
	def is_disable(self):
		return not self.is_enable

	def enable(self):
		self.useraccountcontrol = ["NORMAL_ACCOUNT"]
		self.save()

	def disable(self):
		'''Method to disable User in Active Directory'''
		self.useraccountcontrol = ["NORMAL_ACCOUNT","ACCOUNTDISABLE"]
		self.save()

	def setPassword(self,value):
		self.unicodePwd = password
		self.save()

	@property
	def whenPasswordExpires(self):
		import decimal, datetime

		pwdLastSet = int(self.pwdlastset.raw[0])

		if self.useraccountcontrol.value == '66048':
			return None #Object password not expire

		if (pwdLastSet == 0):
			return 0

		try:
			maxPwdAge = int(self._objects.base.maxpwdage.raw[0])
			mod = int(maxPwdAge) % 4294967296
		except:
			mod = 0
		if mod == 0:
			return None #Domain not expire object passwords

		pwdExpire = decimal.Decimal(pwdLastSet) - decimal.Decimal(maxPwdAge)
		expiryts = int((pwdExpire / 10000000) - 11644473600)

		return datetime.datetime.fromtimestamp(expiryts)

class group(object):
	_type = {
		'objectClass' : [b'top', b'group']
	}

	def addMember(self,object):
		if object.dn == None:
			return None
		try:
			return self._objects._engine._worker.modify(self.dn.value,{'member':[('MODIFY_ADD',[object.dn.value])]})
		except self._exceptions.LDAPEntryAlreadyExistsResult:
			return False

	def delMember(self,object):
		if object.dn == None:
			return None
		try:
			return self._objects._engine._worker.modify(self.dn.value,{'member':[('MODIFY_DELETE',[object.dn.value])]})
		except self._exceptions.LDAPUnwillingToPerformResult:
			return False

	def isMember(self,object):
		if (object.dn != None) and (self.dn != None):
			ingroup = [object.dn.value]
			if object.primarygroupid != None:
				primarygroup = object.primarygroupid.value
				if self.objectsid.value == primarygroup.objectsid.value:
					return True
				ingroup.append(primarygroup.dn.value)
			#Check if object is member or object primarygroup is member
			members = self._objects.search(self._objects.setQuery('or',in_group=ingroup))
			for member in members:
				if (member.dn.value.strip().lower() == self.dn.value.strip().lower()):
					return True
		return False

	@property
	def allMembers(self):
		if self._allmembers != None:
			return self._allmembers
		self._allmembers = []
		primarygroupid=self.objectsid.value.replace(self._objects.base.objectsid.value+'-','')
		members = []
		if self.member != None:
			members.extend(self.member.raw)
		memberlst = self._objects.search(self._objects.setQuery(primarygroupid=primarygroupid))
		if len(memberlst) > 0:
			members.extend(memberlst)
		self._allmembers = members
		return members

class computer(object):
	_type = {
		'objectClass' : [b'top',b'person', b'organizationalPerson', b'user', b'computer']
	}

	@property
	def is_enable(self):
		mod = int(self.useraccountcontrol.value) % 8
		if mod == 0:
			return True
		else:
			return False

	@property
	def is_disable(self):
		return not self.is_enable

	def enable(self):
		self.useraccountcontrol = ["PASSWD_NOTREQD","WORKSTATION_TRUST_ACCOUNT"]
		self.save()

	def disable(self):
		'''Method to disable User in Active Directory'''
		self.useraccountcontrol = ["ACCOUNTDISABLE","PASSWD_NOTREQD","WORKSTATION_TRUST_ACCOUNT"]
		self.save()

class ou(object):
	_type = {
		'objectClass' : [b'top', b'organizationalUnit']
	}



from distutils.core import setup

requires = []

try:
	import ldap3
except:
	requires.append('ldap3')

setup(
	name="pydirectory",
	version="0.1.2",
	description="Python frameWork to managing multiples LDAP services - (gamma)",
	author="Jonas Delgado Mesa",
	author_email="jdelgado@yohnah.net",
	url="https://github.com/Wengex/PyDirectory",
	license="GPLv2",
	packages=[
		'pydirectory',
		'pydirectory.directory',
		'pydirectory.directory.settings',
		'pydirectory.directory.engine',
		'pydirectory.directory.objects',
		'pydirectory.ldap',
		'pydirectory.ldap.settings',
		'pydirectory.ldap.engine',
		'pydirectory.ldap.objects',
		'pydirectory.activedirectory',
		'pydirectory.activedirectory.settings',
		'pydirectory.activedirectory.engine',
		'pydirectory.activedirectory.objects',
	],
	long_description=open('README.txt').read(),
	install_requires = requires
)



from pydirectory.directory.objects import actions
from ldap3 import SUBTREE, BASE

class search (actions.search):
	def object(self,*args,**kwargs):
		return self.__call__(*args,**kwargs)

	def _get(self,query,**kwargs):
		scope = SUBTREE
		if kwargs.get('scope','').upper() == 'BASE':
			scope = BASE

		if kwargs.get("dn",False):
			basedn=kwargs["dn"]
		else:
			basedn=self._objects._engine._settings.basedn

		attributes=['*','+']
		if kwargs.get('attributes',False):
			attributes=kwargs['attributes']


		if (query == None):
			if kwargs.get("dn",False):
				query = '(objectclass=*)'
			else:
				raise self._exceptions.LDAPInvalidFilterError("must be set query filter syntax")

		searchScope = SUBTREE
		c = self._objects._engine._worker
		cookie = None
		while (cookie) or (cookie == None):
			c.search(search_base=basedn,search_filter=query,search_scope = scope, attributes=attributes, paged_size=1000, paged_cookie=cookie)

			if not (c.result.get('result',False) == 0):
				code = c.result.get('result')
				if code == 10:
					raise self._exceptions.LDAPReferrals(c.result['referrals'][0])
				raise self._exceptions.LDAPError(str(c.result))
			cookie = c.result['controls']['1.2.840.113556.1.4.319']['value']['cookie']
			for entry in c.entries:
				entry._response['raw_attributes']['dn'] = [entry._dn]
				try:
					entry._response['raw_attributes']['container'] = [",".join(entry._response['raw_attributes']['dn'][0].split(',')[1:])]
				except TypeError: #python 3.0 compatibility
					entry._response['raw_attributes']['container'] = [b",".join(entry._response['raw_attributes']['dn'][0].split(b',')[1:])]
				self._objectslist.append(entry._response['raw_attributes'])
		return self._objectslist

class get(search):
	def _get(self,*args,**kwargs):
		result= super(get,self)._get(*args,**kwargs)
		if len(result) <= 0:
			raise self._exceptions.ObjectNotExist
		if len(result) > 1:
			raise self._exceptions.MultipleResults
		return result[0]

class new (actions.new):
	def _get(self,data,*args,**kwargs):
		self._objectslist.append(data)
		return self._objectslist[0]



from pydirectory.directory.objects import attributes
from ldap3 import MODIFY_REPLACE


class attribute(attributes.attribute):
	_is_rdn = False
	_operator = MODIFY_REPLACE

class cn(attribute):
	_is_rdn = True

class dn(attribute):
	_is_readonly = True
	_is_rdn = True

class container(attribute):
	_is_rdn = True



from pydirectory.directory.objects import types
from ldap3 import MODIFY_DELETE,MODIFY_REPLACE

class object(types.object):
	def _delete(self):
		if self.dn.value == None:
			return None
		try:
			return self._objects._engine._worker.delete(self.dn.value)
		except self._exceptions.LDAPNoSuchObjectResult:
			return False

	def _save(self):
		self.container._is_modified = False
		if self.dn.value != None:
			modlist = {}
			for key,attr in self._attrs.items():
				if attr._is_modified and (not attr._is_rdn):
					operator = MODIFY_REPLACE
					modlist[key]= [(operator,attr.raw)]
					attr._is_modified = False
			for key in self._drops:
				operator = MODIFY_DELETE
				modlist[key] = [(operator,[])]
			self._drops = []
			if len(modlist) > 0:
				self._objects._engine._worker.modify(self.dn.value,modlist)

			if self.cn._is_modified:
				self._objects._engine._worker.modify_dn(self.dn.value,'CN='+self.cn.value)
				self.dn.update('CN='+self.cn.value+','+','.join(self.dn.value.split(',')[1:]))
				self.cn._is_modified = False

			if self.container._is_modified: #falta por terminar
				self._objects._engine._worker.modify_dn(self.dn.value,'CN='+self.cn.value,new_superior=self.container.value)
				self.dn.update('CN='+self.cn.value+','+self.container.value)
				self.container._is_modified = False
		else:
			attributes = {}
			self.dn.update('CN='+self.cn.value+','+self.container.value)
			for key,attr in self._attrs.items():
				if (key != "dn") and (key != "container"):
					attributes[key]= attr.value
					attr._is_modified = False
			if len(attributes) > 0:
				try:
					self._objects._engine._worker.add(self.dn.value,attributes=attributes)
				except self._exceptions.PyAsn1Error:
					raise self._exceptions.CheckValueAttributes


	def _reset(self):
		if self.dn.value != None:
			obj = self._objects.get(dn=self.dn.value,scope='BASE')
			self._attrs = obj._attrs
		else:
			raise self._exceptions.DNisNone



from pydirectory.ldap.objects import classes
import importlib
class objecttypes(classes.objecttypes):
	def _getobject(self,value):
		types = importlib.import_module("pydirectory.%(type)s.objects.types" % {'type':self._objects._engine._settings.type})
		value.get('objectClass',[]).sort()
		for type in dir(types): #get all classes
			if type.find('__') != 0: #ignore any metaclass. Ej: __whatever__
				try: #check if data in value is same some class
					objclass = getattr(types,type)
					objclass._type.get('objectClass').sort()
					if objclass._type.get('objectClass',[]) == value.get('objectClass',[]):
						return objclass(self._objects,value,objid=value['dn'][0])
				except AttributeError:
					pass
		return types.object(self._objects,value)

class objectslist(classes.objectslist):
	pass

class objects(classes.objects):
	@property
	def base(self):
		obj = self.get(dn=self._engine._settings.basedn.value,scope='base')
		obj._is_readonly = True
		return obj



import importlib

class object(object):
	_objtype = {}
	def __init__(self,objects,data,objid=None,readonly=False):
		if objid == None:
			self._id = id(self)
		else:
			self._id = objid
		self._is_readonly = readonly
		self._objects = objects
		self._exceptions = self._objects._exceptions
		self._attributes = importlib.import_module("pydirectory.%(type)s.objects.attributes" % {'type':self._objects._engine._settings.type})
		self._attrs = {}
		self._initload = True
		self._update(data)
		self._initload = False
		self._drops = []

	def __setitem__(self,key,value):
		if str == bytes:
			if (type(value) == str) or (type(value) == unicode):
				if value.strip() == '':
					value = None
		else:
			if (type(value) == str) or (type(value) == bytes):
				if value.stri() == '':
					value = None
					
		if not self._initload:
			if value == None:
				del self[key]
				return

		def lower_name(s,obj):
			for item in dir(obj):
				if item.lower() == s.lower():
					return item

		try:
			attribute = getattr(self._attributes,str(lower_name(key,self._attributes)))
		except AttributeError:
			attribute = self._attributes.attribute
		if self._initload:
			self._attrs[key.lower()] = attribute(value,self._objects,objid=self._id,modify=False)
		else:
			if not attribute._is_readonly:
				self._attrs[key.lower()] = attribute(value,self._objects,objid=self._id,modify=True)
			else:
				raise self._exceptions.AttributeisReadOnly

	def __getitem__(self,key):
		return self._attrs[key]

	def __delitem__(self,key):
		self._drops.append(key)
		del self._attrs[key]

	def __getattribute__(self,key):

		if key.find('_') != 0:
			if key in self._attrs:
				return self[key]
		try:
			return super(object,self).__getattribute__(key)
		except AttributeError:
			return None

	def __setattr__(self,key,value):
		if key.find('_') == 0:
			super(object,self).__setattr__(key,value)
		else:
			self[key] = value

	def __delattr__(self,key):
		if key.find('_') == 0:
			super(object,self).__delattr__(key)
		del self[key]

	def __len__(self):
		return len(self._attrs)

	def __iter__(self):
		return iter(self._attrs)

	def __dir__(self):
		return self._attrs.keys()

	def _update(self,data):
		for attr,value in data.items():
			self[attr] = value

	def __str__(self):
		return str(self._attrs['dn'])

	def __repr__(self):
		return repr(self._attrs['dn'])

	def items(self):
		return self._attrs.items()

	@property
	def attributes(self):
		return self._attrs

	def delete(self):
		if not self._is_readonly:
			return self._delete()
		else:
			raise self._exceptions.IsReadOnly

	def save(self):
		if not self._is_readonly:
			self._save()
		else:
			raise self._exceptions.IsReadOnly

	def reset(self):
		self._delattr = []
		self._addattr = []
		self._reset()
